/*  MARKOVP.PRG

This is a GAUSS program.  It will implement the estimation and testing
procedures for a Markov switching parameter model as presented in B. Hansen
"The likelihood ratio test under non-standard conditions: Testing the
Markov trend model of GNP."

Written November-December, 1991
Modified May, 1994

written by:

Bruce E. Hansen
Department of Economics
Boston College
Chestnut Hill, MA 02167-3806

(617) 552-3678
Hansen-EC@hermes.bc.edu



The following parameters need to be specified:

y  = dependent variable
x  = independent variables, i.e., constant, lagged y's
iv = 0  to make innovation variance independent of state
iv = 1  to make innovation variance depend upon state
ix = 1|2    Specifies which components of x will depend upon the state.
            Here, ix specifies the first and second components of x.
            For example, if you want the intercept to depend upon the state,
            and the first column of x is a constant, then set "ix = 1".
            To have the intercept and the second AR lag depend upon the state,
            then set "ix = 1|3", if the first column of x is a constant, and
            the third is the second AR lag.
iq = 0  leaves q unconstrained
iq = 1  makes q = 1-p

gn  = number of gridpoints
rep = number of internal Monte Carlo replications used for test

_tnull   = 1  to test the null of one state
_tnull   = 0  to not conduct this test
_nwband  = 4  to set Newey-West bandwidth to 4

_gopt  = 1  to calculate global optimum
_gopt  = 0  to not calculate global optimum

Note:  If _gopt = 1  and _tnull = 0, then one needs to specify starting values
       for the parameters in  "th1".  The order is

(1) Slope coefficients for x in state 0
(2) Standard deviation of error in state 0
(3) Difference in slope coefficients between states
(4) Difference in standard deviation between states (when relevant)
(5) p
(6) q (when relevant).


If _tnull = 1, then it is necessary to specify a grid for the parameters which
depend upon the state, for p, and for q when iq = 0.  The number of gridpoints
must be the same for all regression parameters, but may be different for p or q.
These gridpoints go in the matrices

  "gx", "gp" and "gq"

where gp and gq are vectors, and gx is gn x k1 matrix, gn being number of
gridpoints and k1 being number of parameters which depend upon the state.
That is, each column of gx are the gridpoints for a particular parameter.

Finally, an appropriate likelihood function must be included.  This likelihood
function be a function of all of the above listed variables, even when not
relevant.  The function should return a vector which yields the likelihood
evaluated at each observation, i.e., an nx1 vector.

This program is currently set up to reproduce Hansen's results from
Tables 5 and 6 using "Grid 3"                                            */


/* ***********************************/

output file = "markovp.out" reset;
output off;
n = 135;
load gnp[]=gnp82.dat;
gnp = 100*ln(gnp);
gnp = gnp[2:n+1] - gnp[1:n];
nar = 4;         nn = n-nar;
y = gnp[nar+1:n];

x = ones(nn,1);
j = 1; do while j <= nar;
   x = x~gnp[nar+1-j:n-j];
j = j + 1; endo;

iv = 0;
ix = 1;
iq = 0;
_tnull = 1;
_gopt  = 0;
rep = 1000;
_nwband = 4;

@let th = -.4 .1 0 -.1 -.1 .8 1.5 2.2 .8 ;@

optset;
library optmum;
#include optmum.ext;
_opgtol = .0001;
_opmiter = 500;

gn = 20;
gx = seqa(.1,.1,gn);   @ Note: the columns of gx should equal the rows of ix @

/* *****************************************/

k = cols(x);
kx = rows(ix);
k1 = kx + iv;

if cols(gx) .ne k1;
"";"";
"Columns of matrix  gx  must equal number of parameters which depend upon the";
"state, which are the number of elements in the vector  ix  plus 1 if  iv = 1";
"";
"Program Terminates";
end;
endif;

proc marklike(th);
local sig0,sig1,ff,pp,iz,n,fit,it,z,nks,c,qq,f,zy,p1,f1,ph1,qq0,qq1,p,q1;

  c = sqrt(2*pi);
  sig0 = abs(th[k+1]);
  sig1 = sig0 + abs(th[2*k+2]);
  p = th[2*k+3];
  q1 = 1 - th[2*k+4];
  pp = q1/(1-p+q1);

z = y - x*th[1:k];
qq0 = exp(-(z.^2)./(2*sig0*sig0))./(sig0*c);

z = z - x*th[k+2:2*k+1];
qq1 = exp(-(z.^2)./(2*sig1*sig1))./(sig1*c);
clear z;

n = rows(y);
fit = zeros(n,1);
it = 1;
do until it > n;

  p1 = q1 + (p - q1)*pp;
  f1 = qq1[it]*p1;
  ff = f1 + qq0[it]*(1-p1);
  pp = f1/ff;
  fit[it] = ff;

it = it+1;
endo;
if minc(fit) .> 0;
 f = ln(fit);
else;
 f = -1000;
endif;

retp(-f);
endp;

b1 = zeros(k,1); p=0; q=0;

proc dmclike(th);
local sig0,sig1,ff,pp,iz,n,fit,it,z,nks,c,qq,f,zy,p1,f1,ph1,qq0,qq1,
z2,dqq0,dqq1,dpp,df1,dp1,dff,beta1,beta2,sig02,sig12,qs;

  c = sqrt(2*pi);

  beta1 = th[1:k];
  sig0 = abs(th[k+1]);
  sig1 = sig0;
  if iv==1;
   sig1 = sig1 + abs(b1[k1]);
  endif;
  beta2 = zeros(k,1);
  beta2[ix] = b1[1:kx];
  sig02 = sig0.^2;
  sig12 = sig1.^2;
  qs = 1 - q;

  pp = qs/(1-p+qs);

  z = y - x*beta1;
  z2 = z.^2;
  qq0 = exp(-z2./(2*sig02))./(sig0*c);
  dqq0 = (z.*x)~(z2./sig0 - sig0);
  dqq0 = dqq0.*(qq0./sig02);

  z = z - x*beta2;
  z2 = z.^2;
  qq1 = exp(-z2./(2*sig12))./(sig1*c);
  dqq1 = (z.*x)~(z2./sig1 - sig1);
  dqq1 = dqq1.*(qq1./sig12);

  n = rows(y);
  fit = zeros(1,1+k);
  dpp = zeros(1,1+k);
  it = 1;
  do until it > n;

    p1 = qs + (p - qs)*pp;
    f1 = qq1[it]*p1;
    ff = f1 + qq0[it]*(1-p1);
    pp = f1/ff;

    dp1 = (p-qs).*dpp;
    df1 = dqq1[it,.].*p1 + qq1[it].*dp1;
    dff = (df1 + dqq0[it,.].*(1-p1) - qq0[it].*dp1)./ff;
    dpp = (df1 - f1.*dff)./ff;
    fit = fit + dff;

  it = it+1;
  endo;

retp(-fit);
endp;

pn = 12;
gp = seqa(.1,.075,pn);
qn = pn;
gq = gp;


@  Calculation Under null  @

xxi = invpd(moment(x,0));
b0 = xxi*(x'y);
u = y - x*b0;
v0 = sqrt((u'u)/rows(u));
var0 = xxi*((x.*u)'(x.*u))*xxi;
se0 = sqrt(diag(var0));
b = b0|v0;


proc clike(th);
local th2,ths,qs,lik;
th2 = zeros(k+1,1);
th2[ix] = b1[1:kx];
if iv == 1; th2[k+1] = b1[k1];  endif;
if iq == 1; qs = 1-p; else; qs = q;  endif;
ths = th|th2|p|qs;
lik = marklike(ths);
retp(lik);
endp;

proc mclike(th);
local lik;
lik = sumc(clike(th));
retp(lik);
endp;

proc glike(th);
local lik,p,q,th1,th2,r,th3;
th1 = th[1:k+1];
th2 = zeros(k+1,1);
th2[ix] = th[k+2:k+1+kx];
r = rows(th);
if iq == 1;
 p = th[r];
 p = exp(p)/(1+exp(p));
 q = 1-p;
else;
 p = th[r-1];
 q = th[r];
 p = exp(p)/(1+exp(p));
 q = exp(q)/(1+exp(q));
endif;
th3 = th1|th2|p|q;
lik = marklike(th3);
retp(lik);
endp;

proc mglike(th);
local lik;
lik = sumc(glike(th));
retp(lik);
endp;

b1 = zeros(k1,1);
null = clike(b);
mnull = sumc(null);

if _tnull == 1;

t1 = hsec;
__output = 0;
format 10,4;
"The program is now beginning the grid search over the required parameters.";
"Each parameter setting will be printed on the screen as it comes up.";
"The parameters are printed in this order:";
"     P, Q, Slope Parameters";
"";
"This may take a while, depending upon your computer and the grid you have";
"selected.";
"";

gnn = gn^k1;
cnum = pn*gnn;
if iq == 0;
 cnum = cnum*qn;
endif;
ny = rows(y);
e = rndn(ny+_nwband,rep);
draws = ones(1+_nwband,rep)*(-1000);
c = zeros(cnum,1);
cs = c;
if _gopt == 1;
 beta = zeros(k+3+k1-iq,cnum);
endif;
j = 0;
_opgdprc = &dmclike;

i1 = 1; do while i1 <= rows(gp);
p = gp[i1];

i2 = 1; do while i2 <= rows(gq);
 if iq == 1;
  q = 1-p;
  i2 = rows(gq);
 else;
  q = gq[i2];
 endif;

bs = b;
bi = ones(k1,1);

do while maxc(bi) <= gn;
  b1 = diag(gx[bi,.]);
   j = j + 1;
   {bnew,f,gr,ret} = optmum(&mclike,bs);
   tm = _opitdta[2]*60;
   (p~q~(b1')~_opitdta[1]~tm);
  if _gopt == 1;
   beta[1:k+1,j] = bnew;
   beta[k+2:k+1+k1,j] = b1;
   beta[k+2+k1,j] = p;
   if iq == 0;
    beta[k+3+k1,j] = q;
   endif;
  endif;
  diff = null - clike(bnew);
  diff = diff - meanc(diff);
  se = sqrt(diff'diff);
  diff = diff/se;
  c[j] = mnull - f;
  cs[j] = (mnull-f)/se;
  diffe = zeros(1,rep);
nw = 0; do while nw <= _nwband;
  diffe = diffe + diff'e[1+nw:ny+nw,.];
  draws[1+nw,.] = maxc(draws[1+nw,.]|(diffe./sqrt(1+nw)))';
nw = nw + 1; endo;

  bs = bnew;
  bi[k1] = bi[k1] + 1;
  if bi[k1] > gn;
   bs = b;
   if minc(bi) < gn;
    do until maxc(bi) <= gn;
     ji = maxindc(bi);
     bi[ji] = 1;
     bi[ji-1] = bi[ji-1] + 1;
    endo;
   endif;
  endif;
endo;

i2 = i2 + 1; endo;
i1 = i1 + 1; endo;

mc = maxindc(cs);
m = cs[mc];

draws = draws';
pvalue = meanc(draws .> m);

i = 0; do while i<=_nwband;
  draws[.,i+1] = sortc(draws[.,i+1],1);
i = i + 1; endo;
let cr = .8 .9 .95 .99;
cu = draws[round(cr*rep),.];

t2 = hsec;

endif;

if _gopt == 1;

t3 = hsec;

if _tnull == 1;
 th = beta[.,maxindc(c)];
 p = th[rows(th)-1+iq];
 th[rows(th)-1+iq] = ln(p/(1-p));
 if iq == 0;
  q = th[rows(th)];
  th[rows(th)] = ln(q/(1-q));
 endif;
endif;

__output = 2;
{bnew,f,gr,ret} = optmum(&mglike,th);
lr = sumc(null) - f;
r = rows(bnew);

g = gradp(&glike,bnew);
s = g'g;
h = pinv(hessp(mglike,bnew));
v = h*s*h;

if iq == 1;
 p = bnew[r];
 p = exp(p)/(1+exp(p));
 bopt = bnew[1:r-1]|p;
 v[r,r] = v[r,r]*((p*(1-p))^2);
else;
 p = bnew[r-1];
 q = bnew[r];
 p = exp(p)/(1+exp(p));
 q = exp(q)/(1+exp(q));
 bopt = bnew[1:r-2]|p|q;
 vt = (p*(1-p))|(q*(1-q));
 v[r-1 r,r-1 r] = vt.*v[r-1 r,r-1 r].*(vt');
endif;

se = sqrt(diag(v));
g = cumsumc(g);
stable = sumc(diag((g'g)/s))/rows(g);

t4 = hsec;
endif;

output on;
"Estimation Under Null of Linearity";
" Regression Parameters:"
b0~se0;
"";
"Standard Deviation : " v0;
"";
"Log-Likelihood " mnull;
"";"";
if _gopt == 1;
"Results from estimation and testing of Markov switching model";
"";
"Estimates and standard errors";
"";
"Regression Parameters in State 0";
bopt[1:k]~se[1:k];
"Standard deviation in State 0";
bopt[k+1]~se[k+1];
"";
"Differences in regression Parameters in State 1";
bopt[k+2:k+kx+1]~se[k+2:k+kx+1];
if iv == 1;
 "Standard deviation in State 1";
 bopt[k+kx+2]~se[k+kx+2];
endif;
"P";
r = rows(bopt);
bopt[r-1+iq]~se[r-1+iq];
if iq == 0;
 "Q";
 bopt[r]~se[r];
endif;
"";"";
"Likelihood Value " f;
"Likelihood Ratio " lr;
"2*LR             " 2*lr;
"Nyblom Stability " stable;
"Estimation Time (Hours) " (t4-t3)/360000;
"";"";
endif;
if _tnull == 1;
"Test of Null of One-State Model";
"";
"Grid for p";
pn';
"";
if iq == 0;
 "Grid for Q";
 qn';
 "";
endif;
"Grid for regression Parameters";
gx;
"";
"Internal Monte Carlo Replications" rep;
"";
"Standardized LR Test   " m;
" P-values              ";
seqa(0,1,_nwband+1)~pvalue;
" Upper Criticals";
((1-cr)')|(cu');
"";
if _gopt == 1;
 th = beta[.,mc];
 "Standardized LR maximized at";
 " Regression Parameters";
 th[k+2:k+1+k1];
 "";
 " P, Q";
 th[k+2+k1:k+3+k1-iq];
 "";
endif;
"Computation Time (Hours) " (t2-t1)/360000;
"";"";"";"";
endif;
output off;

